
// (c) COPYRIGHT URI/MIT 1996, 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Dan Holloway <d.holloway@gso.uri.edu> dan

// This is the main CGI program which serves JGOFS data using the DODS DAP. 
//
// DanH 12/20/94

#include "config_jg.h"

static char rcsid[] not_used ={"$Id: jg_dods.cc,v 1.21.4.2 2001/08/21 15:33:11 dan Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>

#include <iostream>
#include <string>

#include <GetOpt.h>

#include "DDS.h"
#include "Sequence.h"
#include "Str.h"
#include "cgi_util.h"
#include "DODSFilter.h"

#include "jgofs.h"
#include "JGSequence.h"
#include "JGStr.h"

extern bool create_dds(DDS &dds, string &obj_path, string &met_path, string &jg_name, string &error, DAS &das);
extern void check_jgofs_path(string &jg_path, string &jg_name);

const string cgi_version = DODS_SERVER_VERSION;

int 
main(int argc, char * argv[])
{
    DDS dds;
    DAS das;
    DODSFilter df(argc, argv);
	
    if (!df.OK()) {
	df.print_usage();
	return 1;
    }

    if (df.version()) {
        df.set_cgi_version(cgi_version);
	df.send_version_info();
	return 0;
    }

    string object_name = df.get_dataset_name();
    string relative_object_path = "";
    size_t slash;
    
    if ( (slash = object_name.find_last_of('/')) != string::npos) {
	relative_object_path = object_name.substr(0,slash+1);
	object_name = object_name.substr(slash+1, object_name.length() - slash+1);
    }

    string jgofsCE = df.get_ce();
    
    df.set_ce(""); // Remove constraint-expression from DodsFilter.
                   // This is required because JGOFS handles the constraint.
    int idx;

    if ( jgofsCE != "" ) {

      while ( (idx = jgofsCE.find("&")) != string::npos)
	  jgofsCE.replace(idx, 1, ',');

      while ( (idx = jgofsCE.find("Level_0.")) != string::npos)
	  jgofsCE.erase(idx, 8);

      object_name += "(" + jgofsCE + ")";

    }


    try {
 
       string method_path = "";
       check_jgofs_path(relative_object_path, method_path);
       
       // Set the filename. 12/22/98 jhrg
       dds.filename(object_name);

       char met[256], par[256];
       string pathname = "";

       // Determine location of the initial file for the
       // JGOFS data object, using the JGOFS object dictionary.
              
       if ( ldctsearch(relative_object_path.c_str(),method_path.c_str(),object_name.c_str(),met,par,1) ) {
	   string objname = par;
	   size_t slash = objname.find_last_of('/');
	   pathname = objname.substr(0, slash+1);
       }

       // Read the ancillary DAS if available.
       df.read_ancillary_das(das,pathname);

       string errMsg;

       // Open the JGOFS data-object and create a DataSet Descriptor Structure
       if (!create_dds(dds, relative_object_path, method_path, object_name, errMsg, das)) {
	   //client error message
	   errMsg = "jg_dods: could not read the DDS information";
	   //httpd error_log message
	   ErrMsgT(errMsg);

	   throw Error(unknown_error, errMsg);
       }

      df.read_ancillary_dds(dds,pathname);

      df.send_data(dds, stdout);
    }

    catch (Error &e) {
      set_mime_text(cout, dods_error, df.get_cgi_version());
      e.print(cout);
      return 1;
    }

    return 0;
}

// $Log: jg_dods.cc,v $
// Revision 1.21.4.2  2001/08/21 15:33:11  dan
// Extensive modifications to handle exceptions properly, and to clean up
// the object dictionary search logic. Added relative path searching to
// locate the object dictionaries, as per request U.S. JGOFS Office.
//
// Revision 1.21.4.1  2001/06/14 21:02:24  dan
// Added DODS_SERVER_VERSION handling to main.
//
// Revision 1.21  2000/10/10 23:36:14  jimg
// Added exception processing code.
// Added throws to the read methods.
// Changed the definition of read to match the dap library.
// Moved CVS Log entries to the end of files.
//
// Revision 1.20  2000/01/20 18:58:00  jimg
// Result of merge with release 3-1-2.
//
// Revision 1.19  1999/11/05 01:17:25  jimg
// Result of merge with release-3-1-0
//
// Revision 1.18.2.2  2000/01/19 23:47:03  dan
// Added default path to the .object dictionary as
// the current working directory.  It will look in
// the 'cwd last for this file.
//
// Revision 1.18.2.1  1999/10/24 18:24:54  dan
// Modified to support new DODS_Type and missing_value
// attribute usage.
//
// Revision 1.17.2.1  1999/06/07 17:29:46  dan
// Changed string class usage 'data()' to 'c_str()'.
//
// Revision 1.17  1999/05/27 15:51:32  dan
// Added logic to find ancillary DDS file relative to the
// top-most file defining a JGOFS data object, as specified
// in the object dictionary.
//
// Revision 1.16  1999/05/25 16:33:55  dan
// Removed extraneous extern definitions.
//
// Revision 1.15  1999/05/05 22:57:19  jimg
// Merged Brent's String --> string changes with Dan's changes
//
// Revision 1.14  1999/01/21 01:46:14  jimg
// Added call to set filename in the DDS. This is required for proper operation
// of the projection functions.
//
// Revision 1.13  1998/03/20 00:30:00  jimg
// Changed calls to the set_mime_*() to match the new parameters
//
// Revision 1.12  1997/08/05 22:45:01  jimg
// Changed type of option_char in `option_char = getopt(...)' from char to
// int. This fixes a bug that shows up on the SGI (where char is unsigned).
//
// Revision 1.11  1997/06/05 17:20:13  jimg
// Rewrote main() so that it uses functions added to cgi_util.cc for the 2.14
// release. These function help to unify the *_dods filters so that they all
// respond more uniformly.
//
// Revision 1.10  1997/03/13 06:12:26  jimg
// Fixed call to create_dds(). Now create_dds() returns an error message if an
// error was detected.
//
// Revision 1.9  1997/03/10 23:05:48  dan
// Modified code to reflect new sub-process changes used when
// compressing the data stream.
//
// Revision 1.7  1996/11/21 23:55:58  jimg
// Changed compression to new compressor function.
//
// Revision 1.6  1996/11/12 17:02:03  jimg
// Removed dead wood.
// Added MIME error returns for certain errors.
//
// Revision 1.5  1996/09/17 00:04:11  jimg
// Misc changes associated with merging branch code to the trunk.
//
// Revision 1.4.2.1.2.1  1996/06/26 22:18:21  jimg
// Dans changes for 2.05.
//
// Revision 1.2  1995/10/11  23:06:27  jimg
// These files are the result of merging Dan's changes into the main
// trunk. However, the archive was messed up in the process, so it may be hard
// to trace back to the original versions of some of these files.
//
// Revision 1.1  1995/09/29  23:31:56  jimg
// Added after much confusion and CVS screw-ups.
//
// Revision 1.1  1995/09/18  22:50:00  jimg
// These files changed as a result of edit Dan made to the jg-dods code so that
// the servers use exiting jgofs methods (i.e., one set of jg-dods servers can
// be used to access N methods, without the linking hack previously used).
//
// Revision 1.3  1995/03/14  02:15:14  jimg
// Modified for the new netio library.
// Removed old code which used C style strings (now it uses C++ Strings).
// Fixed misc bugs.
// Added comments.
//
// Revision 1.2  1995/01/31  21:31:14  dan
// Removed files from repository which have been moved to ../netio
//
