
// (c) COPYRIGHT URI/MIT 1996, 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Dan Holloway <d.holloway@gso.uri.edu> dan

// This file contains functions which read the variables and level
// information from a JGOFS data-object and build the in-memory DDS. These
// functions form the core of the server-side software necessary to extract
// the DDS from a JGOFS data-object file.
//
// It also contains test code which will print the in-memory DDS to
// stdout. It uses both the DDS class as well as the JGOFS object-method
// library. 
//
// jhrg 9/23/94 / dholloway 10/25/94

#include "config_jg.h"

static char rcsid[] not_used ={"$Id: jgdds.cc,v 1.20.4.4 2002/01/15 21:05:08 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <iostream>
#include <strstream.h>
#include <string>

#include "DDS.h"
#include "cgi_util.h"

#include "jgofs.h"		// extern declarations including many globals.
#include "dods-limits.h"
#include "JGByte.h"
#include "JGInt16.h"
#include "JGUInt16.h"
#include "JGInt32.h"
#include "JGUInt32.h"
#include "JGFloat32.h"
#include "JGFloat64.h"
#include "JGSequence.h"
#include "JGStr.h"

// Note that these globals are defined here and declared extern in `jgofs.h'.
// Including that file gives you access to these variables and is required
// when using the JGOFS code. jhrg 12/31/97.

char varName[NVALS][TOKEN];
char varValue[NVALS][VALSIZE];
int varLevel[NVALS];
int nVars, unit;
int varSize = TOKEN;

bool
CreateSequence(int currLvl, int maxLvls, string &error, DAS &das, JGSequence *seq)
{
    if (currLvl <= maxLvls) 
      {
	// Create a new 'Sequence' variable representation, then create a
	// 'BaseType' representation for each variable at this level and add
	// it to current Sequence. When done call CreateSequence with
	// currLvl++.
	ostrstream oss;

	oss << "Level_" << currLvl << ends;
	string name = oss.str();
	oss.freeze(0);

	//JGSequence *seq = new JGSequence(name);
	//seq->set_level(currLvl);

	AttrTable *at;
	AttrType var_attr, var_miss;
	string return_attr = "", dods_var = "";
	char quotedStr[64];
	
	JGByte    *new_byte; 
	JGInt16   *new_int16;
	JGUInt16  *new_uint16;
	JGInt32   *new_int32; 
	JGUInt32  *new_uint32;
	JGFloat32 *new_flt32; 
	JGFloat64 *new_flt64; 
	JGStr     *new_str;

	unsigned char  missing_byte_value;
	short          missing_int16_value;
	unsigned short missing_uint16_value;
	int            missing_int32_value;
	unsigned long  missing_uint32_value; 
	float          missing_float32_value;
	double         missing_float64_value;
	string         missing_str_value;

	for (int i = 0; i < nVars; i++) {
	    if (varLevel[i] == currLvl) {

	      at = das.get_table(varName[i]);

	      if ( at ) {

		return_attr = at->get_attr("DODS_Type");
		downcase(return_attr);
		dods_var = varName[i];

		if (return_attr.find("container") != return_attr.npos)
		  var_attr = Attr_container;
		else if (return_attr.find("byte") != return_attr.npos)
		  var_attr = Attr_byte;
		else if (return_attr.find("uint16") != return_attr.npos)
		  var_attr = Attr_uint16;
		else if (return_attr.find("int16") != return_attr.npos)
		  var_attr = Attr_int16;
		else if (return_attr.find("uint32") != return_attr.npos)
		  var_attr = Attr_uint32;
		else if (return_attr.find("int32") != return_attr.npos)
		  var_attr = Attr_int32;
		else if (return_attr.find("float32") != return_attr.npos)
		  var_attr = Attr_float32;
		else if (return_attr.find("float64") != return_attr.npos)
		  var_attr = Attr_float64;
		else if (return_attr.find("string") != return_attr.npos)
		  var_attr = Attr_string;
		else if (return_attr.find("url") != return_attr.npos)
		  var_attr = Attr_url;
		else 
		  var_attr = Attr_unknown;

		switch (var_attr) {

		  case Attr_byte:
		    new_byte = new JGByte(varName[i]);

		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_byte_value = DODS_UCHAR_MAX;
		    }
		    else if ( var_miss != Attr_byte ) {
		      // Error condition, unless unspecified or 'byte' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type BYTE" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_byte_value = atoi(at->get_attr("missing_value").c_str());
		    }
		    new_byte->set_missing_value(missing_byte_value);
		    seq->add_var(new_byte);
		    break;
		    
		  case Attr_int16:
		    new_int16 = new JGInt16(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_int16_value = DODS_SHRT_MAX;
		      new_int16->set_missing_value(missing_int16_value);
		    }
		    else if ( var_miss != Attr_int16 ) {
		      // Error condition, unless unspecified or 'int16' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type INT16" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_int16_value = atoi(at->get_attr("missing_value").c_str());
		    }
		    new_int16->set_missing_value(missing_int16_value);
		    seq->add_var(new_int16);
		    break;

		  case Attr_uint16:
		    new_uint16 = new JGUInt16(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, set default value.
		      missing_uint16_value = DODS_USHRT_MAX;
		    }
		    else if ( var_miss != Attr_uint16 ) {
		      // Error condition, unless unspecified or 'uint16' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type UINT16" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_uint16_value = atoi(at->get_attr("missing_value").c_str());
		    }
		    new_uint16->set_missing_value(missing_uint16_value);
		    seq->add_var(new_uint16);
		    break;

		  case Attr_int32: 
		    new_int32 = new JGInt32(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_int32_value = DODS_INT_MAX;
		    }
		    else if ( var_miss != Attr_int32 ) {
		      // Error condition, unless unspecified or 'int32' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type INT32" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_int32_value = atoi(at->get_attr("missing_value").c_str());
		    }
		    new_int32->set_missing_value(missing_int32_value);
		    seq->add_var(new_int32);
		    break;

		  case Attr_uint32: 
		    new_uint32 = new JGUInt32(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_uint32_value = DODS_UINT_MAX;
		    }
		    else if ( var_miss != Attr_uint32 ) {
		      // Error condition, unless unspecified or 'uint32' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type UINT32" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_uint32_value = atoi(at->get_attr("missing_value").c_str());
		    }
		    new_uint32->set_missing_value(missing_uint32_value);
		    seq->add_var(new_uint32);
		    break;

		  case Attr_float32:
		    new_flt32 = new JGFloat32(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_float32_value = DODS_FLT_MAX;
		    }
		    else if ( var_miss != Attr_float32 ) {
		      // Error condition, unless unspecified or 'float32' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type FLOAT32" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_float32_value = strtod(at->get_attr("missing_value").c_str(),(char **)FALSE);
		    }
		    new_flt32->set_missing_value(missing_float32_value);
		    seq->add_var(new_flt32);
		    break;

		  case Attr_float64: 
		    new_flt64 = new JGFloat64(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_float64_value = DODS_DBL_MAX;
		    }
		    else if ( var_miss != Attr_float64 ) {
		      // Error condition, unless unspecified or 'float64' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type FLOAT64" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_float64_value = strtod(at->get_attr("missing_value").c_str(),(char **)FALSE);
		    }
		    new_flt64->set_missing_value(missing_float64_value);
		    seq->add_var(new_flt64);
		    break;

		  case Attr_string: 
		    new_str = new JGStr(varName[i]);
		    var_miss = at->get_attr_type("missing_value");
		    if ( var_miss == Attr_unknown ) {
		      // No Missing_Value specification, use default value.
		      missing_str_value = "nd";
		    }
		    else if ( var_miss != Attr_string ) {
		      // Error condition, unless unspecified or 'String' cannot set missing value
		      // to one specified in the ancillary DAS.
		      error += "\"JGOFS/DDS - Variable = " + dods_var + ", DODS_Type attr = " + return_attr +  + ", if provided in ancillary DAS file, missing_value attribute must be type STRING" + "\"";
		      return FALSE;
		    }
		    else {
		      missing_str_value = at->get_attr("missing_value");
		    }
		    new_str->set_missing_value(missing_str_value);
		    seq->add_var(new_str);
		    break;

		  case Attr_url: 
		  case Attr_container:
		    error += "\"JGOFS/DDS - Cannot convert variable `" + (string)varName[i] + "' to type: " + at->get_attr(varName[i]) + ", valid types are Byte, Int16, UInt16, Int32, UInt32, Float32, Float64" + "\"";		
		    return FALSE;

		  case Attr_unknown: 
		    new_str = new JGStr(varName[i]);

		    missing_str_value = "nd";
		    new_str->set_missing_value(missing_str_value);

		    seq->add_var(new_str);
		    break;
		}
	      }
	      else { 
		new_str = new JGStr(varName[i]);
		seq->add_var(new_str);
	      }
	    }
	}
	
	CreateSequence(currLvl+1, maxLvls, error, das, seq); 
	return true;
      }
}

// Given a reference to an instance of class DDS and a filename that refers
// to a JGOFS data object, use the JGOFS data object's method to extract all
// the level and variable information describing the data object's
// hierarchical data representation. Add the variables and level information
// to the instance of DDS.
//
// Returns: false if an error accessing the JGOFS data-object was detected,
// true otherwise. 

bool
create_dds(DDS &dds, string &object_path, string &method_path, string &jg_name, string &error, DAS &das)
{
    char jg_name_c[256], object_path_c[256], method_path_c[256];
    size_t pos;

    strncpy(jg_name_c, jg_name.c_str(), 255);
    jg_name_c[255] = '\0';	// for the truly paranoid...

    strncpy(object_path_c, object_path.c_str(), 255);
    object_path_c[255] = '\0';	// for the truly paranoid...

    strncpy(method_path_c, method_path.c_str(), 255);
    method_path_c[255] = '\0';	// for the truly paranoid...

    // Changed -99 to -NVALS (250). jhrg 11/8/96
    nVars = -(NVALS);
    int nLvls = ljdbopen_(&unit, object_path_c, method_path_c, jg_name_c, varName, &varSize, &nVars);

    if (nLvls < 0) {
	char msg[1024];

        if (ljdberror_(&unit, &msg)) 
	    error += "\"Could not open " + jg_name + ": " + msg + "\"";
	else 
	    error += "\"JGOFS/DDS - could not open object: " +  jg_name + "\"";

	return false;
    }
    
    for (int v = 0; v < nVars; ++v) 
	varLevel[v] = ljdblevel_(&unit, &v);

    //
    // Check for an DAS container named 'Variable-Type'
    //   use the variable type specified in this 
    //   container to convert JGOFS string variables to
    //   this type (where applicable).
    //  
    //   If this container does not exist, and for those 
    //   variables not listed in it, set the default 
    //   variable type to 'string'.

    AttrTable *at;
    
    // When setting the dataset name, use only the part before the
    // constraint. 
    if ((pos = jg_name.find("(")) != jg_name.npos) {
	jg_name = jg_name.substr(0,pos);
    }
    dds.set_dataset_name(name_path(jg_name));

    JGSequence *seq = new JGSequence("Level_0");
    
    CreateSequence(0, nLvls, error, das, seq);
    dds.add_var(seq);

    return true;
}

// $Log: jgdds.cc,v $
// Revision 1.20.4.4  2002/01/15 21:05:08  jimg
// I switched from the numeric constants in values.h/limits.h to those in
// dods-limits.h.
//
// Revision 1.20.4.3  2001/08/21 15:35:18  dan
// Extensive modifications to handle exceptions properly, and to clean up
// the object dictionary search logic. Added relative path searching to
// locate the object dictionaries, as per request U.S. JGOFS Office.
//
// Revision 1.20.4.2  2001/06/06 19:36:17  dan
// To support Sequence changes effective in revision 3.2 of the DAP,
// jg-dods now flattens sequences to a single level.  The jgofs api
// actually treats sequences as flattened tables, rather than as
// a hierarchy of individual levels.  The flattening occurs in the
// jgofs api, read calls to the api only return rows of the table,
// so there is no mechanism in the api to iterate at a particular
// level of the hierarchy, as the current DAP Sequence logic requires.
// This change more accurately reflects the behavior of the jgofs
// api than the previous versions provided, and conforms to the
// DAP Sequence representation.
//
// Revision 1.20.4.1  2001/05/31 19:17:16  dan
// Modified CreateSequence to set the level member variable properly.  This
// is required in the 3.2 DAP Sequence code.
//
// Revision 1.20  2000/10/10 23:36:14  jimg
// Added exception processing code.
// Added throws to the read methods.
// Changed the definition of read to match the dap library.
// Moved CVS Log entries to the end of files.
//
// Revision 1.19  1999/11/05 01:17:26  jimg
// Result of merge with release-3-1-0
//
// Revision 1.18.2.2  1999/10/30 19:46:00  dan
// Fixed bad return type when returning on error condition in new
// attribute handling code.
//
// Revision 1.18.2.1  1999/10/24 18:23:00  dan
// Added attribute handling for DODS_Type and missing_value
// attributes.  If these attributes exist in an ancillary
// DAS their values are used to define the variable's return
// type.  These attributes are defined per variable, their
// values do not cross variable boundaries.
//
// Revision 1.17.4.1  1999/06/07 17:30:47  dan
// Changed string class usage 'data()' to 'c_str()'.
//
// Revision 1.17  1999/05/05 22:57:20  jimg
// Merged Brent's String --> string changes with Dan's changes
//
// Revision 1.16  1999/05/05 04:38:00  brent
// convert String.h code to new stanard lib <string>
//
// Revision 1.15  1998/08/06 16:32:41  jimg
// Fixed misuse of the read(...) member function. Return true if more data
// is to be read, false is if not and error if an error is detected
//
// Revision 1.14  1998/02/19 19:49:17  jimg
// Removed call to Sequence::set_level() since the Sequence::add_var() member
// function now does this.
//
// Revision 1.13  1998/02/05 20:14:39  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.12  1997/08/06 16:53:44  jimg
// Dan fixed nasty string bug where char * was used where String was
// expected.
//
// Revision 1.11  1997/06/05 17:19:27  jimg
// Removed the constraint parameter from create_dds() since it was not used.
// Changed type of parameter `filename' to String. Removed unused code.
//
// Revision 1.10  1997/03/10 23:02:51  dan
// Added Error-handling.  Add calls to new API modules which provide
// access to errors generated by local JGOFS methods.
//
// Revision 1.9  1996/11/09 00:43:18  jimg
// Increased the number of variable that can be processed from 99 to NVALS
// (250).
//
// Revision 1.8  1996/09/17 00:04:14  jimg
// Misc changes associated with merging branch code to the trunk.
//
// Revision 1.7.2.1.2.1  1996/06/26 22:18:25  jimg
// Dans changes for 2.05.
//
// Revision 1.5  1995/10/11  23:56:11  jimg
// Added external symbols defined/declared in jgofs.h to lgofs/ljdb.c. Fixed
// the C constant TOKEN so that it is `40' everywhere.
//
// Revision 1.4  1995/10/11  23:06:30  jimg
// These files are the result of merging Dan's changes into the main
// trunk. However, the archive was messed up in the process, so it may be hard
// to trace back to the original versions of some of these files.
//
// Revision 1.3.2.1  1995/09/18  22:48:07  jimg
// Merged Dan's changes with the trunk version. This will be commited to the
// release-1-0-patches branch and then merged back into the trunk later.
//
// Revision 1.3  1995/07/09  21:40:05  jimg
// Added copyright notice.
//
// Revision 1.2  1995/03/14  02:17:26  jimg
// Modified so that the new netio library is used.
// Modified so that C++ Strings are used in place of C strings.
// Removed old code copied from the dods netcdf sources.
//
// Revision 1.1  1995/01/09  16:04:53  dan
// These files constitute the JGOFS server and client library code
// for use with the DODS api.
//
// Revision 1.1  1994/10/25  9:00:00  dan
// First version of the jgofs software which used libdas++.a.
//
